
#include "OptionsListView.h"

#ifdef CODEWARRIOR
	#pragma mark [Constructor & destructor]
#endif

// Constructor & destructor
// ------------------------

// --------------------------------------------------------------
OptionsListView::OptionsListView
	(
	BRect			frame,
    const char *	name, 
    uint32 			resizeMask, 
  	uint32 			flags
	)
: BView(frame, name, resizeMask, 0)
{
	SetFlags(flags | B_WILL_DRAW | B_FRAME_EVENTS);
	
	SetViewColor(ui_color(B_PANEL_BACKGROUND_COLOR));
	
	m_height = 0;
	m_width = frame.Width();
	m_list = new BList();
	m_scroller = NULL;
}


// --------------------------------------------------------------
OptionsListView::~OptionsListView()
{
	delete m_list;
}

#ifdef CODEWARRIOR
	#pragma mark [Public methods]
#endif

// Public methods
// --------------

// --------------------------------------------------------------
void OptionsListView::TargetedByScrollView
	(
	BScrollView *	scrollview
	)
{
	m_scroller = scrollview;
	SetupScroller();
}


// --------------------------------------------------------------
void OptionsListView::FrameResized
	(
	float width,
	float height
	)
{
	inherited::FrameResized(width, height);
	
	SetupScroller();
}


// --------------------------------------------------------------
void OptionsListView::AddChild
	(
	BView *	view
	)
{
	view->MoveTo(view->Frame().left, view->Frame().top + m_height);
	
	inherited::AddChild(view);

	m_list->AddItem(view);
	m_height = view->Frame().bottom + 1;

	SetupScroller();
}


// --------------------------------------------------------------
void OptionsListView::SetupScroller()
{
	BScrollBar *	sb;
	float			visible_height;

	if ( ! m_scroller )
		return;

	visible_height = m_scroller->Bounds().Height();
	if ( m_height < visible_height )
		ResizeTo(Bounds().Width(), visible_height);

	sb = m_scroller->ScrollBar(B_VERTICAL);
	if ( sb != NULL )
		{
		sb->SetProportion(visible_height / m_height);
		if ( visible_height > m_height )
			sb->SetRange(0, 0);
		else 
			sb->SetRange(0, m_height - visible_height);
		};
}
