
#include <app/Application.h>
#include <InterfaceKit.h>
#include <StorageKit.h>
#include <SupportKit.h>
#include <TranslationKit.h>

#include <stdlib.h>

#include "Sanity.h"
#include "BeSANE.h"
#include "ScannerOptionView.h"
#include "SpinControl.h"

#ifdef CODEWARRIOR
	#pragma mark [Constructor & destructor]
#endif

static const char * get_unit (SANE_Unit unit);

// Constructor & destructor
// ------------------------

// --------------------------------------------------------------
ScannerOptionView::ScannerOptionView
	(
	BRect			frame,
    const char *	name, 
    uint32 			resizeMask, 
  	uint32 			flags, 	
	SANE_Handle 	device,
	int 			option_number
	)
: BView(frame, name, resizeMask, 0)
{
	m_flags 	= flags | B_WILL_DRAW | B_FRAME_EVENTS;
	SetFlags((flags & ~B_WILL_DRAW)| B_FRAME_EVENTS);

	m_device	= device;
	m_opt		= option_number;
	m_desc 		= sane_get_option_descriptor(m_device, m_opt);
	
	m_enabled	= true;
}

// --------------------------------------------------------------
ScannerOptionView::~ScannerOptionView()
{
}

#ifdef CODEWARRIOR
	#pragma mark [Public methods]
#endif

// Public methods
// --------------

status_t ScannerOptionView::Build()
{
	int		nb_values;

	if ( ! m_desc )
		return B_ERROR;

	switch (m_desc->type)
		{
		case SANE_TYPE_INT:
		case SANE_TYPE_FIXED:
			nb_values = m_desc->size / sizeof(SANE_Word);
			break;

		default:
			nb_values = 1;
			break;
		};
		
	switch (m_desc->type)
		{
		case SANE_TYPE_BOOL:
			m_type = BOOLEAN;
			break;
			
		case SANE_TYPE_BUTTON:
			m_type = COMMAND;
			break;
			
		case SANE_TYPE_STRING:
			if (m_desc->constraint_type == SANE_CONSTRAINT_STRING_LIST)
				m_type = STRING_LIST;
			else
				m_type = STRING;
			break;
		
		case SANE_TYPE_INT:
		case SANE_TYPE_FIXED:
			if ( nb_values > 1 )
				{
				m_type = NUMERIC_ARRAY;
				break;
				};
			
			if (m_desc->constraint_type == SANE_CONSTRAINT_WORD_LIST)
				m_type = NUMERIC_LIST;
			else if (m_desc->constraint_type == SANE_CONSTRAINT_RANGE)
				m_type = NUMERIC_RANGE;
			else
				m_type = NUMERIC;
			break;
			
		default:
			m_type = UNKNOWN;
			break;
		};
		
	BString label;
	BRect 	r = Bounds();
	float	w, h;
	bool	enabled;
	
	w = r.Width();
	h = r.Height();

	enabled = SANE_OPTION_IS_SETTABLE(m_desc->cap) && SANE_OPTION_IS_ACTIVE(m_desc->cap);
	label << m_desc->title;
	
	if ( strlen(get_unit(m_desc->unit)) )
		label << " (" << get_unit(m_desc->unit) << ")";

  void * val = malloc(m_desc->size);
  sane_control_option (m_device, m_opt, SANE_ACTION_GET_VALUE, val, 0);
	
	switch (m_type)
		{
		case BOOLEAN:
			{
			BCheckBox * cb;
			cb = new BCheckBox(r, "value", m_desc->title,
							   new BMessage(SET_VALUE_MSG),
								ResizingMode());
			AddChild(cb);
			cb->ResizeToPreferred();
			cb->GetPreferredSize(&w, &h);
			cb->SetFlags(cb->Flags() | m_flags);

			cb->SetValue(*(SANE_Bool *) val);

			break;
			};
			
		case COMMAND:
			{
			BButton * bt;
			bt = new BButton(r, "value", m_desc->title,
									new BMessage(SET_VALUE_MSG));
			AddChild(bt);
			bt->ResizeToPreferred();
			bt->GetPreferredSize(&w, &h);
			bt->SetFlags(bt->Flags() | m_flags);

			break;
			};
			
		case NUMERIC:
			{
			SpinControl * sc;
			r.bottom = r.top + 24;
			sc = new SpinControl(r, "value", label.String(), new BMessage(SET_VALUE_MSG),
									0, 100, 0, 10,
									ResizingMode());
			AddChild(sc);
			sc->SetFlags(sc->Flags() | m_flags);
			
			label << "##";
			
			sc->SetDivider(be_plain_font->StringWidth(label.String()));
			// sc->ResizeToPreferred();
			sc->GetPreferredSize(&w, &h);
			// sc->ResizeTo(w, sc->Frame().Height());
			
			h += 4;

			break;
			};
			
		case NUMERIC_LIST:
			{
			BMenuField *	mf;
			BPopUpMenu *	pum;
			int				i;
			
			pum = new BPopUpMenu("<pick one>");
			pum->SetRadioMode(true);
			
	  		for (i = 0; i < m_desc->constraint.word_list[0]; ++i)
				{
				BMessage * 	msg;
				BString		tmp;
				BMenuItem *	item;
				
				msg = new BMessage(SET_VALUE_MSG);
				
				if (m_desc->type == SANE_TYPE_FIXED)
					{
					float v = SANE_UNFIX(m_desc->constraint.word_list[i+1]);
					tmp << v;
					}
				else
					tmp << m_desc->constraint.word_list[i+1];

				msg->AddInt32("value", m_desc->constraint.word_list[i+1]);

				item = new BMenuItem(tmp.String(), msg);
				if ( (*((SANE_Word *) val)) ==  m_desc->constraint.word_list[i+1])
					item->SetMarked(true);
				pum->AddItem(item);
				};			
			
			mf = new BMenuField(r, "value", label.String(), pum,
								ResizingMode());
			AddChild(mf);

			mf->SetFlags(mf->Flags() | m_flags);
			
			label << "##";
			
			mf->SetDivider(be_plain_font->StringWidth(label.String()));
			mf->ResizeToPreferred();
			mf->GetPreferredSize(&w, &h);
			
			h = 30;

			break;
			}
			
		case NUMERIC_RANGE:
			{
			BSlider * 		slider;
			BString 		min_label, max_label;
				
			BStringView *	sv;
			sv = new BStringView(r, m_desc->name, label.String());
			sv->ResizeToPreferred();
			sv->GetPreferredSize(&w, &h);

			AddChild(sv);

			r.left += w;
								
			slider = new BSlider(r, "value", NULL, // label.String(),
								new BMessage(SET_VALUE_MSG),
								m_desc->constraint.range->min, m_desc->constraint.range->max,
								B_BLOCK_THUMB, // B_TRIANGLE_THUMB,
								ResizingMode());

			AddChild(slider);
			slider->SetFlags(slider->Flags() | m_flags);
		
			if (m_desc->type == SANE_TYPE_FIXED)
				{
				float min = SANE_UNFIX(m_desc->constraint.range->min);
				float max = SANE_UNFIX(m_desc->constraint.range->max);
				min_label << min;
				max_label << max;
				}
			else
				{
				min_label << m_desc->constraint.range->min;
				max_label << m_desc->constraint.range->max;
				};
					
			// min_label << get_unit(m_desc->unit);
			// max_label << get_unit(m_desc->unit);
					
			slider->SetLimitLabels(min_label.String(), max_label.String());
			if ( m_desc->constraint.range->quant )
				{
				slider->SetHashMarkCount(m_desc->constraint.range->quant);
				slider->SetHashMarks(B_HASH_MARKS_BOTTOM);
				};
					
			slider->ResizeToPreferred();
			slider->GetPreferredSize(&w, &h);
			break;
			};
			
		case STRING:
			{
			BTextControl * 	tc;
			
			tc = new BTextControl(r, "value", label.String(), "", // value
									new BMessage(SET_VALUE_MSG),
									ResizingMode());
			AddChild(tc);

			tc->SetFlags(tc->Flags() | m_flags);

			label << "##";
			
			tc->SetDivider(be_plain_font->StringWidth(label.String()));
			tc->ResizeToPreferred();
			tc->GetPreferredSize(&w, &h);

			tc->SetText((char *) val);
			break;
			};
			
		case STRING_LIST:
			{
			BMenuField *	mf;
			BPopUpMenu *	pum;
			int				i;
			
			pum = new BPopUpMenu("<pick one>");
			pum->SetRadioMode(true);
			
			for (i = 0; m_desc->constraint.string_list[i]; i++)
				{
				BMessage * 	msg;
				BMenuItem *	item;
				
				msg = new BMessage(SET_VALUE_MSG);
				msg->AddString("value", m_desc->constraint.string_list[i]);
				item = new BMenuItem(m_desc->constraint.string_list[i], msg);
				if ( strcmp((char *) val, m_desc->constraint.string_list[i]) == 0)
					item->SetMarked(true);
				pum->AddItem(item);
				};			
			
			mf = new BMenuField(r, "value", label.String(), pum,
								ResizingMode());
			AddChild(mf);
			mf->SetFlags(mf->Flags() | m_flags);
			
			label << "##";
			
			mf->SetDivider(be_plain_font->StringWidth(label.String()));
			mf->ResizeToPreferred();
			mf->GetPreferredSize(&w, &h);
			break;
			}
			
		case UNKNOWN:
		default:
			{
			BStringView * sv;
			sv = new BStringView(r, m_desc->name, m_desc->title,
								ResizingMode());
			AddChild(sv);
			sv->SetFlags(sv->Flags() | m_flags);
			sv->ResizeToPreferred();
			sv->GetPreferredSize(&w, &h);
			break;
			};
		};
		

	SetEnabled(enabled);
	ResizeTo(Bounds().Width(), h);

	free(val);

	return B_OK;
}


// --------------------------------------------------------------
void ScannerOptionView::AttachedToWindow()
{
	BInvoker * 		invoker;
	BMenuField *	menufield;
	BView *	child;

	child = ChildAt(0);
	while ( child )
		{
		invoker = dynamic_cast<BInvoker *>(child);
		if ( invoker )
			invoker->SetTarget(this);
			
		menufield = dynamic_cast<BMenuField *>(child);
		if ( menufield )
			menufield->Menu()->SetTargetForItems(this);
		
		child = child->NextSibling();
		};

	if ( Parent() )
		SetViewColor(Parent()->ViewColor());
	// SetViewColor((rand() % 256), (rand() % 256), (rand() % 256));
		
	inherited::AttachedToWindow();
}


// --------------------------------------------------------------
void ScannerOptionView::SetEnabled
	(
	bool	on
	)
{
	BControl * 	control;
	BView *		child;

	child = ChildAt(0);
	while ( child )
		{
		control = dynamic_cast<BControl *>(child);
		if ( control )
			control->SetEnabled(on);
		child = child->NextSibling();
		};

	m_enabled = on;
}


// --------------------------------------------------------------
void ScannerOptionView::MessageReceived
	(
	BMessage *	msg
	)
{
	switch (msg->what)
		{
		case B_SIMPLE_DATA:
		case BROWSE_FILE_MSG:
		case BROWSE_DIR_MSG:
			{
			// msg->PrintToStream();

			if ( ! m_enabled )
				break;

		   	entry_ref ref;

  			// Look for a ref in the message
   			if( msg->FindRef("refs", &ref) == B_OK )
				{
				BEntry 	entry(&ref);
				BPath	path;
				
				if ( entry.InitCheck() != B_OK )
					break;
				
				entry.GetPath(&path);
					
				BTextControl * control = dynamic_cast<BTextControl *>(FindView("value"));
				if ( control )
					{
					control->SetText(path.Path());
					Looper()->PostMessage(SET_VALUE_MSG, this);
					};
				break;
	   			};

			// Call inherited if we didn't handle the message
   			inherited::MessageReceived(msg);
			break;
			};
			
		case SET_VALUE_MSG:
			{
			SANE_Status	status;
			SANE_Word	word_value;
			SANE_Bool 	bool_value;
			SANE_Int 	info;
			void *		new_value;
			
			// msg->PrintToStream();
			
			new_value = NULL;
			switch ( m_type )
				{
				case BOOLEAN:
					{
					BControl * control = dynamic_cast<BControl *>(FindView("value"));
					if ( ! control )
						break;
					bool_value = control->Value();
					new_value = &bool_value;
					break;
					};

				case STRING:
					{
					BTextControl * tc = dynamic_cast<BTextControl *>(FindView("value"));
					if ( tc )
						new_value = (void *) tc->Text();
					break;
					};
					
				case STRING_LIST:
					msg->FindString("value", (const char **) &new_value);
					break;
					
				case NUMERIC_LIST:
					if ( msg->FindInt32("value", (int32 *) &word_value) == B_OK )
						new_value = &word_value;
					break;
					
				default:
					break;
				};
					
			if ( new_value == NULL )
				break;
							
			status = sane_control_option(m_device, m_opt, SANE_ACTION_SET_VALUE,
										new_value, &info);
			if ( status != SANE_STATUS_GOOD )
			 	fprintf (stderr, "sane_control_option(%s, SET_VALUE): %s\n",
						m_desc->name, sane_strstatus (status));
			break;
			};
			
		default:	
			inherited::MessageReceived(msg);
		};
}


static const char * get_unit (SANE_Unit unit)
{
  switch (unit)
    {
    case SANE_UNIT_PIXEL:
      return "pixels";
    case SANE_UNIT_BIT:
      return "bit";
    case SANE_UNIT_MM:
      return "mm";
    case SANE_UNIT_DPI:
      return "dpi";
    case SANE_UNIT_PERCENT:
      return "%";
    case SANE_UNIT_MICROSECOND:
      return "us";
    default:
      return "";
    };
}



